@@ -32,6 +32,7 @@ struct TableDesc {
   size_t key_size;  // sizes are in bytes
   size_t leaf_size;
   size_t max_entries;
+  int flags;
   std::string key_desc;
   std::string leaf_desc;
   llvm::Function *key_sscanf;
@@ -39,6 +40,7 @@ struct TableDesc {
   llvm::Function *key_snprintf;
   llvm::Function *leaf_snprintf;
   bool is_shared;
+  bool is_extern;
 };
 
 }  // namespace ebpf
